import numpy as np

def BiotSavart( curve: np.array, curveTangent: np.array, fieldPoints: np.array ):
    pointFieldStrength = np.zeros( (len(curve[0,:]), len( fieldPoints[:,0] ), len( fieldPoints[0,:] ) ) )

    s = range( len(curve[0,:]) )

    for i in s:
        pointDistances =  fieldPoints - curve[:,i].reshape(-1,1)

        pointNorms = np.linalg.norm(pointDistances, axis=0)
        pointNormsCubed = np.power( pointNorms, 3)

        crossProduct = ( np.cross(curveTangent[:,i], pointDistances.T ) ).T
        pointFieldStrength[i,:,:] = crossProduct / pointNormsCubed

    v = np.trapz( pointFieldStrength, s , axis=0)
    return v


import numpy as np
from ClosedDerivatives import closedFirstGradient, closedSecondGradient


def KappaBinormal( curve: np.array, meshpoints: np.array ):
    # Finding the unit tangent vector
    dc = closedFirstGradient(curve, meshpoints)
    curveTangent = np.divide( dc, np.linalg.norm(dc, axis=0) )

    # Finding the unit normal vector
    d2c = closedSecondGradient(curve, meshpoints)
    curveNormal = np.divide( d2c, np.linalg.norm(d2c, axis=0) )

    # Using the unit tangent and unit normal vectors to find the binormal vector
    curveBinormal = np.cross(curveTangent,curveNormal, axisa=0, axisb=0, axisc=0)

    # Defining Kappa
    dTdt = closedFirstGradient(curveTangent,meshpoints)
    Kappa = np.divide( np.linalg.norm( dTdt,  axis=0 ), np.linalg.norm( dc, axis=0 ) )
    KBApprox = np.multiply(Kappa, curveBinormal)

    return KBApprox


from filamentlib.BiotSavart import BiotSavart
from filamentlib.RegularBiotSavart import RegularBiotSavart
from filamentlib.KappaBinormal import KappaBinormal
from filamentlib.EvolveKappaBinormal import EvolveKappaBinormal
from filamentlib.EvolveRegularBiotSavart import EvolveRegularBiotSavart


import numpy as np

# Calculating the closed first derivative of a bunch of 3D points
def closedFirstGradient(f: np.array, t: np.array = None) -> np.array:
    if t is None:
        t = np.linspace(0,1,f.shape[1],endpoint=False)

    fPlusOne = np.roll( f, -1, axis=1 )
    fMinusOne = np.roll( f, 1, axis=1 )
    
    # Define the time step for every value of t
    h = np.roll(t, -1) - t
    h[-1] = h[0]

    num = fPlusOne - fMinusOne
    den = 2 * h
    grad = np.divide(num,den)

    return grad


# Calculating the closed second derivative of a bunch of 3D points
def closedSecondGradient(f: np.array, t: np.array = None):
    if t is None:
        t = np.linspace(0,1,f.shape[1],endpoint=False)

    fPlusOne = np.roll( f, -1, axis=1 )
    fMinusOne = np.roll( f, 1, axis=1 )

    # Define the time step for every value of t
    h = np.roll(t, -1) - t
    h[-1] = h[0]

    num = fPlusOne - 2*f + fMinusOne
    den = np.power(h,2)
    grad = num/den

    return grad


import numpy as np
from KappaBinormal import KappaBinormal
from scipy.integrate import solve_ivp

# Functions to update a curve over time
# Euler's method on Kappa Binormal
def EulerKappaBinormal( curve:np.array, meshpoints: np.array, deltaStep: float, maxSteps: int):
    curves = np.zeros( (maxSteps + 1,curve.shape[0],curve.shape[1]) )
    curves[0,:,:] = curve

    for i in range(1,maxSteps+1):
        curves[i,:,:] = curves[i-1, :, :] + deltaStep * KappaBinormal(curves[i-1,:,:], meshpoints)
    return curves

def solve_ivp_KappaBinormal(curve:np.array, meshpoints:np.array) -> np.array:
    curve = curve.reshape( (3, meshpoints.shape[0]) )
    KBApprox = KappaBinormal( curve, meshpoints )

    return KBApprox.flatten()

def EvolveKappaBinormal( curve:np.array, meshpoints: np.array, tspan: list, method: str = 'RK45' ) -> np.array: 
    # Prepare the function and the initial condition to be put through scipy.integrate.solve_ivp
    y0 = curve.flatten()
    KBEqn = lambda t,y: solve_ivp_KappaBinormal( y, meshpoints )

    # solve the system
    soln = solve_ivp( KBEqn, tspan, y0, method=method )

    # Get the simulation curves to be in a standard format
    curves = np.zeros( (soln.y.shape[1], curve.shape[0], curve.shape[1]) )
    for i in range( soln.y.shape[1] ):
        curves[i,:,:] = soln.y[:,i].reshape( curve.shape )

    return curves


from setuptools import find_packages, setup


setup(
    name='filamentlib',
    version='0.1.0',
    description='Simulate the evolution of any given vortex filament over time',
    packages=find_packages(),
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Daniel Krawciw',
    author_email='dkrawciw@mines.edu',
    license='MIT',
    python_requires='>=3.10',
    install_requires=['numpy >= 1.26','scipy >= 1.13.0']
)


import numpy as np
from scipy.integrate import solve_ivp

from ClosedDerivatives import closedFirstGradient
from RegularBiotSavart import RegularBiotSavart

# Method to flatten and unflatten inputted lists of curve points
# Definitely really inefficient but I haven't thought of a better way to use scipy's solve_ivp method without flattening the matrix
def solve_ivp_BiotSavart( curve:np.array, meshpoints:np.array, eps:float ) -> np.array:
        curve = curve.reshape( (3,meshpoints.shape[0]) )
        BSStep = RegularBiotSavart( curve, closedFirstGradient(curve, meshpoints), curve, eps )

        return BSStep.flatten()

# Method to 
def EvolveRegularBiotSavart( curve:np.array, meshpoints: np.array, eps:float, tspan: list, method: str = 'RK45' ) -> np.array:
        y0 = curve.flatten()
        BSEqn = lambda t,y: solve_ivp_BiotSavart( y, meshpoints, eps )

        soln = solve_ivp( BSEqn, tspan, y0, method )

        # Get the simulation curves to be in a standard format
        curves = np.zeros( (soln.y.shape[1], curve.shape[0], curve.shape[1]) )
        for i in range( soln.y.shape[1] ):
            curves[i,:,:] = soln.y[:,i].reshape( curve.shape )

        return curves